from __future__ import annotations
from typing import Optional, Union

from abc import ABC, abstractmethod
import contextlib
import math
import numpy as np
import torch
import luz

__all__ = [
    "Scorer",
    "CrossValidation",
    "Holdout",
]

Device = Union[str, torch.device]


class Scorer(ABC):
    @abstractmethod
    def score(
        self,
        learner: luz.Learner,
        dataset: luz.Dataset,
        device: Union[torch.device, str],
    ) -> tuple[torch.nn.Module, float]:
        """Learn a model and estimate its future performance."""
        pass


class CrossValidation(Scorer):
    def __init__(
        self,
        num_folds: int,
        val_fraction: Optional[int] = None,
        fold_seed: Optional[int] = None,
        shuffle: Optional[bool] = True,
    ) -> None:
        """Object which scores a learning algorithm using cross validation.

        Parameters
        ----------
        num_folds
            Number of cross validation folds.
        val_fraction
            Fraction of data to use as a validation set, by default None.
        fold_seed
            Seed for random fold split, by default None.
        shuffle
            If True, shuffle dataset before splitting into folds; by default True.
        """
        self.num_folds = num_folds
        self.val_fraction = val_fraction
        self.fold_seed = fold_seed
        self.shuffle = shuffle

    def score(
        self,
        learner: luz.Learner,
        dataset: luz.Dataset,
        device: Optional[Device] = "cpu",
    ) -> luz.Score:
        """Learn a model and estimate its future performance using cross validation.

        Parameters
        ----------
        learner
            Learning algorithm to be scored.
        dataset
            Dataset to use for scoring.
        device
            Device to use for scoring, by default "cpu".

        Returns
        -------
        torch.nn.Module
            Learned model.
        float
            Cross-validation score.
        """
        if self.fold_seed is None:
            cm = contextlib.nullcontext()
        else:
            cm = luz.temporary_seed(self.fold_seed)

        test_losses = []

        points_per_fold = math.ceil(len(dataset) / self.num_folds)
        fold_lengths = np.repeat(points_per_fold, self.num_folds)
        fold_lengths[-1] -= fold_lengths.sum() - len(dataset)

        folds = dataset.split(fold_lengths, self.shuffle)

        for i, test_dataset in enumerate(folds):
            train_dataset = luz.ConcatDataset(folds[:i] + folds[i + 1 :])
            train_dataset.use_collate(dataset._collate)
            if self.val_fraction is None:
                val_dataset = None
            else:
                n = len(train_dataset)
                n_val = round(self.val_fraction * n)

                train_dataset, val_dataset = train_dataset.split(
                    [n - n_val, n_val], self.shuffle
                )

            with cm:
                model, test_loss = learner.learn(
                    train_dataset=train_dataset,
                    val_dataset=val_dataset,
                    test_dataset=test_dataset,
                    device=device,
                )

            test_losses.append(test_loss)

        score = sum(test_losses) / self.num_folds

        model = learner.learn(train_dataset=dataset, device=device)

        return model, score


class Holdout(Scorer):
    def __init__(
        self, test_fraction: float, val_fraction: Optional[float] = None
    ) -> None:
        """Object which scores a learning algorithm using the holdout method.

        Parameters
        ----------
        test_fraction
            Fraction of data to use as a test set for scoring.
        val_fraction
            Fraction of data to use as a validation set, by default None.
        """
        self.test_fraction = test_fraction
        self.val_fraction = val_fraction

    def score(
        self,
        learner: luz.Learner,
        dataset: luz.Dataset,
        device: Optional[Device] = "cpu",
    ) -> luz.Score:
        """Learn a model and estimate its future performance using the holdout method.

        Parameters
        ----------
        learner
            Learning algorithm to be scored.
        dataset
            Dataset to use for scoring.
        device
            Device to use for scoring, by default "cpu".

        Returns
        -------
        torch.nn.Module
            Learned model
        float
            Holdout score.
        """
        n = len(dataset)
        n_test = round(self.test_fraction * n)

        if self.val_fraction is not None:
            n_val = round(self.val_fraction * n)
            train_dataset, val_dataset, test_dataset = dataset.split(
                [n - n_val - n_test, n_val, n_test]
            )

            return learner.learn(
                train_dataset=train_dataset,
                val_dataset=val_dataset,
                test_dataset=test_dataset,
                device=device,
            )
        else:
            train_dataset, test_dataset = dataset.split([n - n_test, n_test])

            return learner.learn(
                train_dataset=train_dataset, test_dataset=test_dataset, device=device
            )



from luz.callbacks import *
from luz.datasets import *
from luz.learner import *
from luz.loggers import *
from luz.metrics import *
from luz.modules import *
from luz.runner import *
from luz.scorers import *
from luz.transforms import *
from luz.tuners import *
from luz.utils import *



from __future__ import annotations
from typing import Any, Callable, Optional, Union

import functools
import luz
import pyee
import torch

__all__ = ["Event", "Runner", "State"]

Device = Union[str, torch.device]
Model = Union[torch.nn.Module, dict[str, torch.nn.Module]]


class Event(pyee.EventEmitter):
    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name
        self.count = 0

    def __call__(self, state: State) -> None:
        self.count += 1
        self.emit(self.name, state)

    def reset(self, state: State) -> None:
        self.count = 0

    def attach(
        self,
        f: Callable[..., Any],
        once: Optional[int] = None,
        every: Optional[int] = None,
        filter: Optional[Callable[[int], bool]] = None,
    ) -> None:
        if once is not None:

            @functools.wraps(f)
            def g(*args, **kwargs):
                if self.count == once:
                    f(*args, **kwargs)

            self.on(self.name, g)

        elif every is not None:

            @functools.wraps(f)
            def g(*args, **kwargs):
                if self.count % every == 0:
                    f(*args, **kwargs)

            self.on(self.name, g)

        elif filter is not None:

            @functools.wraps(f)
            def g(*args, **kwargs):
                if filter(self.count):
                    f(*args, **kwargs)

            self.on(self.name, g)
        else:
            self.on(self.name, f)


class State:
    def __init__(self, **kwargs: Any) -> None:
        """Runner state.

        Parameters
        ----------
        **kwargs
            Additional state variables.
        """
        self.data = None
        self.device = None
        self.epoch = None
        self.handlers = []
        self.history = []
        self.ind = None
        self.loader = None
        self.loss = None
        self.max_epochs = None
        self.metrics = {}
        self.model = None
        self.output = None
        self.transform = None

        self.update(**kwargs)

    def update(self, **kwargs: Any) -> None:
        for k, v in kwargs.items():
            setattr(self, k, v)


class Runner:
    def __init__(
        self,
        run_batch: Callable[[State, luz.Data], Any],
        model: torch.nn.Module,
        loader: torch.utils.data.DataLoader,
        max_epochs: int,
        metrics: Optional[dict[str, luz.Metric]] = None,
    ) -> None:
        """Runner.

        Parameters
        ----------
        run_batch
            Function to process batch of data.
        model
            Model applied to data batches.
        loader
            Data loader.
        max_epochs
            Maximum number of epochs.
        metrics
            Metrics computed while running through data, by default None.
        """
        self.RUNNER_STARTED = Event("RUNNER_STARTED")
        self.EPOCH_STARTED = Event("EPOCH_STARTED")
        self.BATCH_STARTED = Event("BATCH_STARTED")
        self.BATCH_ENDED = Event("BATCH_ENDED")
        self.EPOCH_ENDED = Event("EPOCH_ENDED")
        self.RUNNER_ENDED = Event("RUNNER_ENDED")

        self.run_batch = run_batch

        self.state = State(model=model, loader=loader, max_epochs=max_epochs)

        if metrics is not None:
            for name, metric in metrics.items():
                metric.attach(self, name)

    def run(
        self,
        device: Optional[Device] = "cpu",
    ) -> None:
        """Run through data loader.

        Parameters
        ----------
        device
            Device used by runner, by default "cpu".
        """
        self.state.update(device=device)

        self.EPOCH_STARTED.count = 0
        self.EPOCH_ENDED.count = 0

        self.RUNNER_STARTED(self.state)

        for epoch in range(self.state.max_epochs):
            self.state.update(epoch=epoch)

            self.BATCH_STARTED.count = 0
            self.BATCH_ENDED.count = 0

            self.EPOCH_STARTED(self.state)

            for i, data in enumerate(self.state.loader):
                data = data.to(self.state.device)

                self.state.update(ind=i)

                self.BATCH_STARTED(self.state)

                self.run_batch(self.state, data)

                if torch.cuda.is_available():
                    torch.cuda.synchronize()

                self.BATCH_ENDED(self.state)

            self.EPOCH_ENDED(self.state)

        self.RUNNER_ENDED(self.state)


# def supervised_trainer(model, loader, metrics=None, input_keys=None, target_key=None):
#     if input_keys is None:
#         input_keys = ["x"]
#     if target_key is None:
#         target_key = "y"

#     def run_batch_supervised_train(state, batch):
#         state.output = state.model(*(getattr(batch, k) for k in input_keys))
#         state.loss = state.criterion(state.output, getattr(batch, target_key))

#         state.loss.backward()
#         state.optimizer.step()
#         state.optimizer.zero_grad()

#     return Runner(run_batch_supervised_train, model, loader, metrics)



from __future__ import annotations
from typing import Any, Iterable, Optional, Type, Union

import ast
import itertools
import luz
import numpy as np
import operator
import re
import torch

__all__ = ["GridSearch", "RandomSearch", "StopTuning", "Trial"]

Device = Union[str, torch.device]


class Trial(dict):
    def __getattr__(self, key: str) -> Any:
        if key.startswith("__") and key.endswith("__"):
            return dict.__getattr__(self, key)

        return self[key]


class StopTuning(Exception):
    pass


class Tuner:
    def __init__(
        self, learner: luz.Learner, scorer: luz.Scorer, num_iterations: int
    ) -> None:
        self.learner = learner
        self.scorer = scorer
        self.num_iterations = num_iterations
        # self.seed = seed

        self.trials = []
        self.scores = []

    def tune(
        self,
        dataset: luz.Dataset,
        device: Optional[Device] = "cpu",
        **hparams: Union[Sample, Choose, Pin, Conditional],
    ) -> torch.nn.Module:
        """Learn a model based on a given dataset.

        Parameters
        ----------
        train_dataset
            Training dataset used to learn a model.
        device
            Device to use for learning, by default "cpu".

        Returns
        -------
        torch.nn.Module
            Learned model.
        """
        self.trials = []
        self.scores = []

        while True:
            try:
                trial = self.get_trial(**hparams)
                score = self._objective(trial, dataset, device)

                self.trials.append(trial)
                self.scores.append(score)
            except StopTuning:
                break

        return self.best_trial.model

    @property
    def best_trial(self) -> Trial:
        return self.trials[np.argmin(self.scores)]

    def sample(
        self,
        lower: Union[int, float],
        upper: Union[int, float],
        dtype: Union[Type[int], Type[float]],
    ) -> Sample:
        return Sample(lower, upper, dtype)

    def choose(self, *choices: Any) -> Choose:
        return Choose(choices)

    def pin(self, equation: str) -> Pin:
        return Pin(equation)

    def conditional(self, condition: str, if_true: Any, if_false: Any) -> Conditional:
        return Conditional(condition, if_true, if_false)

    def _objective(self, trial: Trial, dataset: luz.Dataset, device: Device) -> float:
        self.learner.hparams = trial

        model, score = self.scorer.score(self.learner, dataset, device)

        trial.learner = self.learner
        trial.model = model
        trial.score = score

        return score


class RandomSearch(Tuner):
    def get_trial(self, **hparams: Union[Sample, Choose, Pin, Conditional]) -> Trial:
        if len(self.trials) == self.num_iterations:
            raise StopTuning

        d = {}

        for k, v in hparams.items():
            if isinstance(v, Sample) and v.dtype == int:
                d[k] = int(np.random.randint(low=v.lower, high=v.upper))
            elif isinstance(v, Sample) and v.dtype == float:
                d[k] = np.random.uniform(low=v.lower, high=v.upper)
            elif isinstance(v, Choose):
                d[k] = np.random.choice(a=v.choices)
            elif isinstance(v, Pin) or isinstance(v, Conditional):
                d[k] = v(**d)

        return Trial(**d)


class GridSearch(Tuner):
    def __init__(self, learner: luz.Learner, scorer: luz.Scorer) -> None:
        super().__init__(learner, scorer, None)

    def get_trial(self, **hparams: Union[Sample, Choose, Pin, Conditional]) -> Trial:
        choices = [v for v in hparams.values() if isinstance(v, Choose)]
        d = {}
        grid = itertools.product(*[c.choices for c in choices])

        for g in grid:
            t = dict(zip(hparams.keys(), g))

            if t not in self.trials:
                d.update(**t)
                break

        if d == {}:
            raise StopTuning

        for k, v in hparams.items():
            if isinstance(v, Pin) or isinstance(v, Conditional):
                d[k] = v(**d)

        return Trial(**d)


class BayesianTuner(Tuner):
    pass
    # def get_trial(self, hparams, trials, scores):
    #     pass # FIGURE OUT HOW TO WRITE THIS


# ---------- AUXILLIARY CLASSES ---------- #


class Sample:
    def __init__(
        self,
        lower: Union[int, float],
        upper: Union[int, float],
        dtype: Union[Type[int], Type[float]],
    ) -> None:
        self.lower = lower
        self.upper = upper
        self.dtype = dtype


class Choose:
    def __init__(self, choices: Iterable[Any]) -> None:
        self.choices = tuple(choices)


class Pin:
    def __init__(self, equation: str) -> None:
        self.equation = equation

    def __call__(self, **kwargs: Any) -> Any:
        # NOTE: this sorting should ensure that longer
        # variable names are substituted first, which
        # should prevent partial variable name clobbering
        pattern = "|".join(sorted(kwargs, key=len)[::-1])

        def replacement(m):
            return str(kwargs[re.escape(m.group(0))])

        expression = re.sub(pattern=pattern, repl=replacement, string=self.equation)

        # parse pin equation string into operators and
        # evaluate operators to obtain the pin value
        return _evaluate_expression(expression)


class Conditional:
    def __init__(self, condition: str, if_true: Any, if_false: Any) -> None:
        self.pin = Pin(condition)
        self.if_true = if_true
        self.if_false = if_false

    def __call__(self, **kwargs: Any) -> Any:
        if self.pin(**kwargs):
            return self.if_true
        return self.if_false


# from https://stackoverflow.com/a/9558001

ops = {
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: operator.mul,
    ast.Div: operator.truediv,
    ast.Pow: operator.pow,
    ast.BitXor: operator.xor,
    ast.USub: operator.neg,
    ast.FloorDiv: operator.floordiv,
    ast.Eq: operator.eq,
    ast.NotEq: operator.ne,
    ast.Lt: operator.lt,
    ast.LtE: operator.le,
    ast.Gt: operator.gt,
    ast.GtE: operator.ge,
    ast.Is: operator.is_,
    ast.IsNot: operator.is_not,
}


def _evaluate_expression(expression: str) -> Any:
    node = ast.parse(expression, mode="eval").body

    return _evaluate_operators(node)


def _evaluate_operators(node) -> Any:
    # FIXME: type annotate `node`
    # <number>
    if isinstance(node, ast.Num):
        return node.n
    # <left> <operator> <right>
    elif isinstance(node, ast.BinOp):
        return ops[type(node.op)](
            _evaluate_operators(node.left),
            _evaluate_operators(node.right),
        )
    # <operator> <operand> e.g., -1
    elif isinstance(node, ast.UnaryOp):
        return ops[type(node.op)](_evaluate_operators(node.operand))
    elif isinstance(node, ast.Compare):
        # from https://github.com/danthedeckie/simpleeval/blob/master/simpleeval.py
        right = _evaluate_operators(node.left)
        to_return = True
        for operation, comp in zip(node.ops, node.comparators):
            if not to_return:
                break
            left = right
            right = _evaluate_operators(comp)
            to_return = ops[type(operation)](left, right)
        return to_return
    else:
        raise TypeError(node)



from __future__ import annotations
from typing import Any, Callable, Optional, Tuple, Union

import collections
import contextlib
import functools
import math
import numpy as np
import os
import pathlib
import random
import torch


__all__ = [
    "adjacency",
    "aggregate_mask",
    "batchwise_edge_mean",
    "batchwise_edge_sum",
    "batchwise_mask",
    "batchwise_node_mean",
    "batchwise_node_sum",
    "dot_product_attention",
    "expand_path",
    "in_degree",
    "int_length",
    "masked_softmax",
    "mkdir_safe",
    "memoize",
    "nodewise_edge_mean",
    "nodewise_edge_sum",
    "nodewise_mask",
    "out_degree",
    "remove_self_loops",
    "set_seed",
    "temporary_seed",
]

Device = Union[str, torch.device]
Func = Callable[..., Any]


def adjacency(
    edge_index: torch.Tensor, device: Optional[Device] = "cpu"
) -> torch.Tensor:
    """Convert edge indices to adjacency matrix.

    Parameters
    ----------
    edge_index
        Edge index tensor.
        Shape: :math:`(2,N_{edges})`

    Returns
    -------
    torch.Tensor
        Adjacency matrix.
        Shape: :math:`(N_{nodes},N_{nodes})`
    """
    n = edge_index.max() + 1
    A = torch.zeros((n, n), device=device).long()
    A[tuple(edge_index)] = 1
    return A


def aggregate_mask(
    groups: torch.Tensor,
    num_groups: int,
    num_items: int,
    mean: Optional[bool] = False,
    device: Optional[Device] = "cpu",
) -> torch.Tensor:
    """Create a mask to aggregate items into groups.

    Parameters
    ----------
    groups
        Tensor of group indices.
    num_groups
        Number of groups.
    num_items
        Number of items.
    mean
        If True, mask aggregates by averaging rather than summing, by default False.
    device
        Device on which mask will be created, by default "cpu".

    Returns
    -------
    torch.Tensor
        Mask to aggregate items into groups.
    """
    M = torch.zeros(num_groups, num_items, device=device)
    M[groups, torch.arange(num_items)] = 1

    if mean:
        M = torch.nn.functional.normalize(M, p=1, dim=1)

    return M


# def aggregate_edges(
#     edges: torch.Tensor,
#     edge_index: torch.Tensor,
#     batch: torch.Tensor,
#     weights: Optional[torch.Tensor] = None,
# ) -> torch.Tensor:
#     if weights is None:
#         if nodewise:
#             weights = nodewise_mask(edge_index, device=edges.device)
#         else:
#             weights = batchwise_mask(batch, edge_index, device=edges.device)

#         if mean:
#             weights = torch.nn.functional.normalize(weights, p=1, dim=1)

#     return weights @ edges


def batchwise_edge_mean(
    edges: torch.Tensor,
    edge_index: torch.Tensor,
    batch: torch.Tensor,
    device="cpu",
) -> torch.Tensor:
    M = batchwise_mask(batch, edge_index, mean=True, device=device)

    return M @ edges


def batchwise_edge_sum(
    edges: torch.Tensor,
    edge_index: torch.Tensor,
    batch: torch.Tensor,
    device="cpu",
) -> torch.Tensor:
    M = batchwise_mask(batch, edge_index, device=device)

    return M @ edges


def batchwise_mask(
    batch: torch.Tensor,
    edge_index: Optional[torch.Tensor] = None,
    mean: Optional[bool] = False,
    device: Optional[Device] = "cpu",
) -> torch.Tensor:
    """Create a mask for batchwise aggregation of graph nodes or edges.

    Parameters
    ----------
    batch
        Tensor of batch indices.
        Shape: :math:`(N_{nodes},)`
    edge_index
        Tensor of edge indices, by default None.
        Shape: :math:`(2,N_{edges})`
    mean
        If True, mask aggregates by averaging rather than summing, by default False.
    device
        Device on which mask will be created, by default "cpu".

    Returns
    -------
    torch.Tensor
        Mask for batchwise aggregation.
        Shape: :math:`(N_{batch},N_{nodes})` if `edge_index = None`
    """
    if edge_index is not None:
        # masking for edge aggregation
        _, N_e = edge_index.shape
        s, r = edge_index
        return aggregate_mask(
            batch[r], batch[r].max() + 1, N_e, mean=mean, device=device
        )
    else:
        # masking for node aggregation
        N_v = len(batch)
        return aggregate_mask(batch, batch.max() + 1, N_v, mean=mean, device=device)


def batchwise_node_mean(
    nodes: torch.Tensor, batch: torch.Tensor, device="cpu"
) -> torch.Tensor:
    M = batchwise_mask(batch, mean=True, device=device)

    return M @ nodes


def batchwise_node_sum(
    nodes: torch.Tensor, batch: torch.Tensor, device="cpu"
) -> torch.Tensor:
    M = batchwise_mask(batch, device=device)

    return M @ nodes


def dot_product_attention(
    query: torch.Tensor, key: torch.Tensor, mask: Optional[torch.Tensor] = None
) -> torch.Tensor:
    """Compute scaled dot product attention.

    Parameters
    ----------
    query
        Query vectors.
        Shape: :math:`(N_{queries},d_q)`
    key
        Key vectors.
        Shape: :math:`(N_{keys},d_q)`
    mask
        Mask tensor to ignore query-key pairs, by default None.
        Shape: :math:`(N_{queries},N_{keys})`

    Returns
    -------
    torch.Tensor
        Scaled dot product attention between each query and key vector.
        Shape: :math:`(N_{queries},N_{keys})`
    """

    _, d = key.shape

    pre_attn = query @ key.transpose(0, 1) / math.sqrt(d)

    return masked_softmax(pre_attn, mask, dim=1)


def expand_path(path: str, dir: Optional[str] = None) -> str:
    p = pathlib.Path(path).expanduser()
    if dir is not None:
        p = pathlib.Path(dir).joinpath(p)
    return str(p.expanduser().resolve())


def in_degree(adjacency: torch.Tensor) -> torch.Tensor:
    """Compute in-degrees of nodes in a graph.

    Parameters
    ----------
    adjacency
        Adjacency matrix.
        Shape: :math:`(N_{nodes}, N_{nodes})`

    Returns
    -------
    torch.Tensor
        Nodewise in-degree tensor.
        Shape: :math:`(N_{nodes},)`
    """
    return adjacency.sum(dim=0)


def int_length(n: int) -> int:
    # from https://stackoverflow.com/a/2189827
    if n > 0:
        return int(math.log10(n)) + 1
    elif n == 0:
        return 1
    else:
        # +1 if you don't count the '-'
        return int(math.log10(-n)) + 2


def masked_softmax(
    x: torch.Tensor, mask: torch.Tensor, dim: Optional[int] = 1
) -> torch.Tensor:
    """Compute softmax of a tensor using a mask.

    Parameters
    ----------
    x
        Argument of softmax.
    mask
        Mask tensor.
    dim
        Dimension along which softmax will be computed.

    Returns
    -------
    torch.Tensor
        Masked softmax of `x`.
    """
    X = (mask * x).masked_fill(mask == 0, -float("inf"))
    return torch.softmax(X, dim)


def mkdir_safe(directory: str) -> None:
    with contextlib.suppress(FileExistsError):
        os.makedirs(directory)


def memoize(func: Func) -> Func:
    func.cache = collections.OrderedDict()

    @functools.wraps(func)
    def memoized(*args, **kwargs):
        k = (args, frozenset(kwargs.items()))
        if k not in func.cache:
            func.cache[k] = func(*args, **kwargs)

        return func.cache[k]

    return memoized


def nodewise_edge_mean(
    edges: torch.Tensor, edge_index: torch.Tensor, device="cpu"
) -> torch.Tensor:
    M = nodewise_mask(edge_index, mean=True, device=device)

    return M @ edges


def nodewise_edge_sum(
    edges: torch.Tensor, edge_index: torch.Tensor, device: Optional[Device] = "cpu"
) -> torch.Tensor:
    """Sum all incoming edge features for each node.

    Parameters
    ----------
    edges
        Edge features.
        Shape: :math:`(N_{edges},d_e)`
    edge_index
        Ege indices.
        Shape: :math:`(2,N_{edges})`
    device
        Device on which mask will be created, by default "cpu".

    Returns
    -------
    torch.Tensor
        Tensor of summed incoming edges for each node.
        Shape: :math:`(N_{nodes},d_e)`
    """
    M = nodewise_mask(edge_index, device=device)

    return M @ edges


def nodewise_mask(
    edge_index: torch.Tensor,
    mean: Optional[bool] = False,
    device: Optional[Device] = "cpu",
) -> torch.Tensor:
    """Create a mask for nodewise aggregation of incoming graph edges.

    Parameters
    ----------
    edge_index
        Edge indices.
        Shape: :math:`(2,N_{edges})`
    mean
        If True, mask aggregates by averaging rather than summing, by default False.
    device
        Device on which mask will be created, by default "cpu".

    Returns
    -------
    torch.Tensor
        Mask for nodewise edge aggregation.
        Shape: :math:`(N_{nodes},N_{edges})`
    """
    _, N_e = edge_index.shape
    s, r = edge_index
    return aggregate_mask(r, r.max() + 1, N_e, mean=mean, device=device)


def out_degree(adjacency: torch.Tensor) -> torch.Tensor:
    """Compute out-degrees of nodes in a graph.

    Parameters
    ----------
    adjacency
        Adjacency matrix.
        Shape: :math:`(N_{nodes},N_{nodes})`

    Returns
    -------
    torch.Tensor
        Nodewise out-degree tensor.
        Shape: :math:`(N_{nodes},)`
    """
    return adjacency.sum(dim=1)


def remove_self_loops(
    edge_index: torch.Tensor, edges: Optional[torch.Tensor] = None
) -> Union[torch.Tensor, Tuple[torch.Tensor]]:
    s, r = edge_index
    mask = s != r

    if edges is None:
        return edge_index[:, mask]

    return edge_index[:, mask], edges[mask, :]


def set_seed(seed: int) -> None:
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


@contextlib.contextmanager
def temporary_seed(seed: int) -> None:
    # adapted from https://stackoverflow.com/a/49557127
    random_state = random.getstate()
    np_state = np.random.get_state()
    torch_state = torch.random.get_rng_state()

    set_seed(seed)

    try:
        yield
    finally:
        random.setstate(random_state)
        np.random.set_state(np_state)
        torch.random.set_rng_state(torch_state)



"""

Contains callback objects which perform various actions during the training process.

"""

from __future__ import annotations
from typing import Any, Optional, Union

from abc import ABC, abstractmethod
import luz
import pathlib
import torch

__all__ = [
    "Callback",
    "Checkpoint",
    "LogMetrics",
    "Progress",
    "Run",
    "UpdateState",
]

Device = Union[str, torch.device]
Path = Optional[Union[str, pathlib.Path]]


class Callback(ABC):
    @abstractmethod
    def __call__(self, state: luz.State) -> Any:
        """Execute callback."""
        pass


class Checkpoint(Callback):
    def __init__(self, model_name: str, save_dir: Optional[Path] = None) -> None:
        self.model_name = model_name

        if save_dir is None:
            save_dir = "."

        self.save_dir = luz.expand_path(path=save_dir)

        luz.mkdir_safe(self.save_dir)

    def __call__(self, state: luz.State) -> None:
        """Execute callback."""
        model = state.model
        epoch = state.epoch

        save_path = pathlib.Path(self.save_dir, f"{self.model_name}_{epoch}.pth.tar")
        torch.save(obj=model.state_dict(), f=save_path)


class LogMetrics(Callback):
    def __init__(self, *loggers: luz.Logger) -> None:
        self.loggers = loggers
        if len(self.loggers) == 0:
            self.loggers = [luz.ConsoleLogger()]

    def __call__(self, state: luz.State) -> None:
        epoch = state.epoch
        metrics = state.metrics
        msg = "\n".join([f"[Epoch {epoch + 1}] {k}: {m}" for k, m in metrics.items()])
        for logger in self.loggers:
            logger.log(msg)


class Progress(Callback):
    def __call__(self, state: luz.State) -> None:
        """Execute callback."""
        e = state.epoch + 1
        b = state.ind + 1
        print(
            f"Epoch {e}/{state.max_epochs}, batch {b}/{len(state.loader)}",
            end="\r",
            flush=True,
        )


class Run(Callback):
    def __init__(self, runner: luz.Runner, device: Device) -> None:
        self.runner = runner
        self.device = device

    def __call__(self, state: luz.State) -> None:
        """Execute callback."""
        self.runner.run(self.device)


class UpdateState(Callback):
    def __init__(self, **kwargs: Any) -> None:
        self.kwargs = kwargs

    def __call__(self, state: luz.State) -> None:
        """Execute callback."""
        state.update(**self.kwargs)



"""

Custom PyTorch modules.

"""
from __future__ import annotations
from typing import Callable, Optional

import luz
import torch

__all__ = [
    "AdditiveAttention",
    "AdditiveNodeAttention",
    "ApplyFunction",
    "AverageGraphPool",
    "Concatenate",
    "Dense",
    "DenseRNN",
    "DotProductAttention",
    "ElmanRNN",
    "EdgeAggregateLocal",
    "EdgeAggregateLocalHead",
    "EdgeAggregateGlobal",
    "EdgeAggregateGlobalHead",
    "GraphConv",
    "GraphConvAttention",
    "GraphNetwork",
    "MaskedSoftmax",
    "NodeAggregate",
]

Activation = Callable[[torch.Tensor], torch.Tensor]


class AdditiveAttention(torch.nn.Module):
    """Additive attention, from https://arxiv.org/abs/1409.0473."""

    def __init__(
        self, d: int, d_attn: int, activation: Optional[Activation] = None
    ) -> None:
        """Additive attention, from https://arxiv.org/abs/1409.0473.

        Parameters
        ----------
        d
            Feature length.
        d_attn
            Attention vector length.
        activation
            Activation function, by default None.
        """
        super().__init__()
        self.concat = Concatenate(dim=1)
        if activation is None:
            activation = torch.nn.Tanh()
        self.W = Dense(2 * d, 2 * d_attn, bias=False, activation=activation)
        self.v = torch.nn.Linear(2 * d_attn, 1, bias=False)

        self.ms = MaskedSoftmax(dim=1)

    def forward(
        self, s: torch.Tensor, h: torch.Tensor, mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        s
            Shape: :math:`(N,d)`
        h
            Shape: :math:`(N,d)`
        mask
            Mask tensor, by default None.

        Returns
        -------
        torch.Tensor
            Output tensor.
            Shape: :math:`(1,N)`
        """
        X = self.W(self.concat(s, h))
        pre_attn = self.v(X).t()

        return self.ms(pre_attn, mask)


class AdditiveNodeAttention(torch.nn.Module):
    def __init__(
        self, d: int, d_attn: int, activation: Optional[Activation] = None
    ) -> None:
        """Additive node attention on graphs. From https://arxiv.org/abs/1710.10903.

        Parameters
        ----------
        d
            Node feature length.
        d_attn
            Attention vector length.
        activation
            Activation function, by default None.
        """
        super().__init__()
        self.attn = AdditiveAttention(d, d_attn, activation)

    def forward(
        self,
        nodes: torch.Tensor,
        edge_index: torch.Tensor,
    ) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        nodes
            Node features.
            Shape: :math:`(N_{nodes},d_v)`
        edge_index
            Edge index tensor.
            Shape: :math:`(2,N_{edges})`

        Returns
        -------
        torch.Tensor
            Output tensor.
            Shape: :math:`(N_{edges},N_{nodes})`
        """
        mask = luz.nodewise_mask(edge_index, device=nodes.device)
        s, r = edge_index
        return self.attn(nodes[s], nodes[r], mask)


class ApplyFunction(torch.nn.Module):
    def __init__(self, f: Callable[[torch.Tensor], torch.Tensor]) -> None:
        self.f = f

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        return self.f(x)


class AverageGraphPool(torch.nn.Module):
    def __init__(self, num_clusters: int) -> None:
        super().__init__()
        self.num_clusters = num_clusters

    def forward(
        self,
        nodes: torch.Tensor,
        edges: torch.Tensor,
        edge_index: torch.Tensor,
        batch: torch.Tensor,
        assignment: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Pool graph by average node clustering.

        Parameters
        ----------
        nodes
            Node features.
            Shape: :math:`(N_{nodes},d_v)`
        edges
            Edge features.
            Shape: :math:`(N_{edges},d_e)`
        edge_index
            Edge index tensor.
            Shape: :math:`(2,N_{edges})`
        batch
            Nodewise batch tensor.
            Shape: :math:`(N_{nodes},)`
        assignment
            Soft cluster assignment tensor.
            Shape: :math:`(N_{nodes},N_{clusters})`

        Returns
        -------
        torch.Tensor
            Pooled node features.
            Shape: :math:`(N_{nodes}',d_v)`
        torch.Tensor
            Pooled edge features.
            Shape: :math:`(N_{edges}',d_e)`
        torch.Tensor
            Pooled edge index tensor.
            Shape: :math:`(2,N_{edges}')`
        """
        M = luz.batchwise_mask(batch).repeat_interleave(self.num_clusters, dim=0).T
        A = assignment.tile(batch.max() + 1)
        _, cluster = luz.masked_softmax(A, M).argmax(dim=1).unique(return_inverse=True)

        # FIXME: compute new node features using internal node AND edge features?
        # intuition: benzene functional group feature should depend on edge features,
        # else losing internal bonding info to distinguish e.g. benzene from cyclohexane
        N_v, _ = nodes.shape
        M = luz.aggregate_mask(
            cluster, cluster.max() + 1, N_v, mean=True, device=nodes.device
        )
        coarse_nodes = M @ nodes

        coarse_edge_index, coarse_edges = luz.remove_self_loops(
            cluster[edge_index], edges
        )

        N_e, _ = coarse_edges.shape
        if N_e > 0:
            coarse_edge_index, indices = coarse_edge_index.unique(
                dim=1, return_inverse=True
            )
            M = luz.aggregate_mask(
                indices, indices.max() + 1, N_e, mean=True, device=edges.device
            )
            coarse_edges = M @ coarse_edges

        return coarse_nodes, coarse_edges, coarse_edge_index


class Concatenate(torch.nn.Module):
    def __init__(self, dim: Optional[int] = 0) -> None:
        """Concatenate tensors along a given dimension.

        Parameters
        ----------
        dim
            Concenation dimension, by default 0
        """
        super().__init__()
        self.dim = dim

    def forward(self, *tensors: torch.Tensor) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        *args
            Input tensors.
            Shape: :math:`(N,*)`

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        return torch.cat(tensors, dim=self.dim)


class Dense(torch.nn.Module):
    def __init__(
        self,
        *features: int,
        bias: Optional[bool] = True,
        activation: Optional[Activation] = None,
    ) -> None:
        """Dense feed-forward neural network.

        Parameters
        ----------
        *features
            Number of features at each layer.
        bias
            If False, each layer will not learn an additive bias; by default True.
        activation
            Activation function.
        """
        super().__init__()

        layers = []

        for n_in, n_out in zip(features, features[1:]):
            layers.append(torch.nn.Linear(n_in, n_out, bias=bias))
            if activation is not None:
                layers.append(activation)

        self.seq = torch.nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        x
            Input tensor.
            Shape: :math:`(N, *, H_{in})`

        Returns
        -------
        torch.Tensor
            Output tensor.
            Shape: :math:`(N, *, H_{out})`
        """
        return self.seq(x)


class DenseRNN(torch.nn.Module):
    def __init__(self, input_size: int, hidden_size: int, output_size: int) -> None:
        super().__init__()

        self.hidden_size = hidden_size
        self.hidden = self._init_hidden()

        self.i2h = torch.nn.Linear(input_size + hidden_size, hidden_size)
        self.i2o = torch.nn.Linear(input_size + hidden_size, output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        self.hidden = self._init_hidden()

        for i in x.transpose(0, 1):
            combined = torch.cat((i, self.hidden), 1)
            self.hidden = self.i2h(combined)
            output = self.i2o(combined)

        return output

    def _init_hidden(self) -> torch.Tensor:
        return torch.zeros(1, self.hidden_size)


class DotProductAttention(torch.nn.Module):
    """Scaled dot product attention."""

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        query
            Query vectors.
            Shape: :math:`(N_{queries},d_q)`
        key
            Key vectors.
            Shape: :math:`(N_{keys},d_q)`
        mask
            Mask tensor to ignore query-key pairs, by default None.
            Shape: :math:`(N_{queries},N_{keys})`

        Returns
        -------
        torch.Tensor
            Scaled dot product attention between each query and key vector.
            Shape: :math:`(N_{queries},N_{keys})`
        """
        return luz.dot_product_attention(query, key, mask)


class EdgeAggregateLocalHead(torch.nn.Module):
    def __init__(
        self,
        d_v: int,
        d_e: int,
        d_u: int,
        d_attn: int,
        nodewise: Optional[bool] = True,
    ) -> None:
        """Aggregates graph edges using attention.

        Parameters
        ----------
        d_v
            Node feature length.
        d_e
            Edge feature length.
        d_u
            Global feature length.
        d_attn
            Attention vector length.
        """
        super().__init__()
        act = torch.nn.LeakyReLU()
        self.query = Dense(d_v + d_u, d_attn, activation=act)
        self.key = Dense(d_v + d_u, d_attn, activation=act)
        self.value = Dense(d_e + d_u, d_attn, activation=act)

        self.concat = Concatenate(dim=1)
        self.attn = DotProductAttention()

        self.lin = Dense(d_attn, d_e, activation=act)

    def forward(
        self,
        nodes: torch.Tensor,
        edges: torch.Tensor,
        edge_index: torch.Tensor,
        u: torch.Tensor,
        batch: torch.Tensor,
    ) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        nodes
            Node features.
            Shape: :math:`(N_{nodes},d_v)`
        edges
            Edge features.
            Shape: :math:`(N_{edges},d_e)`
        edge_index
            Edge index tensor.
            Shape: :math:`(2,N_{edges})`
        u
            Global features.
            Shape: :math:`(N_{batch},d_u)`
        batch
            Nodewise batch tensor.
            Shape: :math:`(N_{nodes},)`

        Returns
        -------
        torch.Tensor
            Output tensor.
            Shape: :math:`(N_{nodes},d_e)`
        """
        mask = luz.nodewise_mask(edge_index, device=edges.device)

        s, r = edge_index
        q = self.query(self.concat(nodes, u[batch]))
        k = self.key(self.concat(nodes[s], u[batch[s]]))
        v = self.value(self.concat(edges, u[batch[s]]))

        attn = self.attn(q, k, mask)

        x = attn @ v

        return self.lin(x)


class EdgeAggregateGlobalHead(torch.nn.Module):
    def __init__(self, d_v: int, d_e: int, d_u: int, d_attn: int) -> None:
        """Aggregates graph edges using attention.

        Parameters
        ----------
        d_v
            Node feature length.
        d_e
            Edge feature length.
        d_u
            Global feature length.
        d_attn
            Attention vector length.
        """
        super().__init__()
        act = torch.nn.LeakyReLU()
        self.query = Dense(d_e + d_u, d_attn, activation=act)
        self.key = Dense(d_u, d_attn, activation=act)
        self.value = Dense(d_e + d_u, d_attn, activation=act)

        self.concat = Concatenate(dim=1)
        self.attn = DotProductAttention()

        self.lin = Dense(d_attn, d_e, activation=act)

    def forward(
        self,
        nodes: torch.Tensor,
        edges: torch.Tensor,
        edge_index: torch.Tensor,
        u: torch.Tensor,
        batch: torch.Tensor,
    ) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        nodes
            Node features.
            Shape: :math:`(N_{nodes},d_v)`
        edges
            Edge features.
            Shape: :math:`(N_{edges},d_e)`
        edge_index
            Edge index tensor.
            Shape: :math:`(2,N_{edges})`
        u
            Global features.
            Shape: :math:`(N_{batch},d_u)`
        batch
            Nodewise batch tensor.
            Shape: :math:`(N_{nodes},)`

        Returns
        -------
        torch.Tensor
            Output tensor.
            Shape: :math:`(N_{batch},d_e)`
        """
        mask = luz.batchwise_mask(batch, edge_index, device=edges.device).t()

        s, r = edge_index
        q = self.query(self.concat(edges, u[batch[s]]))
        k = self.key(self.concat(u))
        v = self.value(self.concat(edges, u[batch[s]]))

        attn = self.attn(q, k, mask).t()

        x = attn @ v

        return self.lin(x)


class ElmanRNN(torch.nn.Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers=None,
        nonlinearity=None,
        bias=None,
        batch_first=None,
        dropout=None,
        bidirectional=None,
        h0=None,
    ):
        super().__init__()

        self.hidden_size = hidden_size

        kwargs = {}
        if num_layers:
            kwargs["num_layers"] = num_layers
        if nonlinearity:
            kwargs["nonlinearity"] = nonlinearity
        if bias:
            kwargs["bias"] = bias
        if batch_first:
            kwargs["batch_first"] = batch_first
        if dropout:
            kwargs["dropout"] = dropout
        if bidirectional:
            kwargs["bidirectional"] = bidirectional

        self.h0 = h0
        self.rnn = torch.nn.RNN(
            input_size=input_size, hidden_size=hidden_size, **kwargs
        )

    def forward(self, x):
        if self.h0:
            output, hidden = self.rnn.forward(input=x, h0=self.h0)
        else:
            output, hidden = self.rnn.forward(input=x)
        return output


class GraphConv(torch.nn.Module):
    def __init__(self, d_v: int, activation: Activation) -> None:
        """Graph convolutional network from https://arxiv.org/abs/1609.02907.

        Parameters
        ----------
        d_v
            Node feature length.
        activation
            Activation function.
        """
        super().__init__()
        self.lin = Dense(d_v, d_v, activation=activation)

    def forward(self, nodes: torch.Tensor, edge_index: torch.Tensor) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        nodes
            Node features.
            Shape: :math:`(N_{nodes},d_v)`
        edge_index
            Edge indices.
            Shape: :math:`(2,N_{edges})`

        Returns
        -------
        torch.Tensor
            Output tensor.
            Shape: :math:`(N_{nodes},d_v)`
        """
        N_v, _ = nodes.shape
        A = luz.adjacency(edge_index, device=nodes.device) + torch.eye(
            N_v, device=nodes.device
        )
        d = luz.in_degree(A).pow(-0.5)
        d.masked_fill(d == float("inf"), 0)
        D = torch.diag(d)

        return self.lin(D @ A @ D @ nodes)


class GraphConvAttention(torch.nn.Module):
    def __init__(self, d_v: int, activation: Optional[Activation] = None) -> None:
        """Compute node attention weights using graph convolutional network.

        Parameters
        ----------
        d_v
            Node feature length.
        activation
            Activation function.
        """
        super().__init__()
        if activation is None:
            activation = torch.nn.Identity()

        self.gcn = GraphConv(d_v, torch.nn.Identity())
        self.lin = Dense(d_v, 1, activation=activation)

    def forward(
        self, nodes: torch.Tensor, edge_index: torch.Tensor, batch: torch.Tensor
    ) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        nodes
            Node features.
            Shape: :math:`(N_{nodes},d_v)`
        edge_index
            Edge indices.
            Shape: :math:`(2,N_{edges})`
        batch
            Batch indices.
            Shape: :math:`(N_{nodes},)`

        Returns
        -------
        torch.Tensor
            Attention weights.
            Shape: :math:`(N_{batch},N_{nodes})`
        """
        pre_attn = self.lin(self.gcn(nodes, edge_index)).t()
        M = luz.batchwise_mask(batch, device=nodes.device)
        attn = luz.masked_softmax(pre_attn, M, dim=1)

        return attn


class GraphNetwork(torch.nn.Module):
    """Graph Network from https://arxiv.org/abs/1806.01261."""

    def __init__(
        self,
        edge_model: Optional[torch.nn.Module] = None,
        node_model: Optional[torch.nn.Module] = None,
        global_model: Optional[torch.nn.Module] = None,
        num_layers: Optional[int] = 1,
    ) -> None:
        """[summary]

        Parameters
        ----------
        edge_model
            Edge update network, by default None.
        node_model
            Node update network, by default None.
        global_model
            Global update network, by default None.
        num_layers
            Number of passes, by default 1.
        """
        super().__init__()
        self.edge_model = edge_model
        self.node_model = node_model
        self.global_model = global_model
        self.num_layers = num_layers

    def forward(
        self,
        nodes: torch.Tensor,
        edge_index: torch.Tensor,
        edges: Optional[torch.Tensor] = None,
        u: Optional[torch.Tensor] = None,
        batch: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        nodes
            Node features.
            Shape: :math:`(N_{nodes},d_v)`
        edge_index
            Edge index tensor.
            Shape: :math:`(2,N_{edges})`
        edges
            Edge features, by default None.
            Shape: :math:`(N_{edges},d_e)`
        u
            Global features, by default None.
            Shape: :math:`(N_{batch},d_u)`
        batch
            Nodewise batch tensor, by default None.
            Shape: :math:`(N_{nodes},)`

        Returns
        -------
        torch.Tensor
            Output node feature tensor.
            Shape: :math:`(N_{nodes},d_v)`
        torch.Tensor
            Output edge index tensor.
            Shape: :math:`(2,N_{edges})`
        torch.Tensor
            Output edge feature tensor.
            Shape: :math:`(N_{edges},d_e)`
        torch.Tensor
            Output global feature tensor.
            Shape: :math:`(N_{batch},d_u)`
        torch.Tensor
            Output batch tensor.
            Shape: :math:`(N_{nodes},)`
        """
        if batch is None:
            N_v, *_ = nodes.shape
            batch = torch.zeros((N_v,), dtype=torch.long)

        for _ in range(self.num_layers):
            if self.edge_model is not None:
                edges = self.edge_model(nodes, edge_index, edges, u, batch).reshape(
                    edges.shape
                )

            if self.node_model is not None:
                nodes = self.node_model(nodes, edge_index, edges, u, batch).reshape(
                    nodes.shape
                )

            if self.global_model is not None:
                u = self.global_model(nodes, edge_index, edges, u, batch).reshape(
                    u.shape
                )

        return nodes, edge_index, edges, u, batch


class MaskedSoftmax(torch.nn.Module):
    """Compute softmax of a tensor using a mask."""

    def __init__(self, dim: Optional[int] = None) -> None:
        super().__init__()
        self.dim = dim

    def forward(
        self, x: torch.Tensor, mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        x
            Argument of softmax.
        mask
            Mask tensor with the same shape as `x`, by default None.

        Returns
        -------
        torch.Tensor
            Masked softmax of `x`.
        """
        if mask is None:
            return torch.softmax(x, self.dim)
        return luz.masked_softmax(x, mask, self.dim)


class EdgeAggregateLocal(torch.nn.Module):
    def __init__(
        self,
        d_v: int,
        d_e: int,
        d_u: int,
        d_attn: int,
        num_heads: Optional[int] = 1,
    ) -> None:
        """Aggregates graph edges using multihead attention.

        Parameters
        ----------
        d_v
            Node feature length.
        d_e
            Edge feature length.
        d_u
            Global feature length.
        d_attn
            Attention vector length.
        num_heads
            Number of attention heads.
        """
        super().__init__()
        self.concat = luz.Concatenate(dim=1)
        self.heads = torch.nn.ModuleList()
        self.gates = torch.nn.ModuleList()

        for _ in range(num_heads):
            h = EdgeAggregateLocalHead(d_v, d_e, d_u, d_attn)
            g = Dense(d_v + d_u, 1, activation=torch.nn.LeakyReLU())

            self.heads.append(h)
            self.gates.append(g)

    def forward(
        self,
        nodes: torch.Tensor,
        edges: torch.Tensor,
        edge_index: torch.Tensor,
        u: torch.Tensor,
        batch: torch.Tensor,
    ) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        nodes
            Node features.
            Shape: :math:`(N_{nodes},d_v)`
        edge_index
            Edge index tensor.
            Shape: :math:`(2,N_{edges})`
        edges
            Edge features, by default None.
            Shape: :math:`(N_{edges},d_e)`
        u
            Global features, by default None.
            Shape: :math:`(N_{batch},d_u)`
        batch
            Nodewise batch tensor, by default None.
            Shape: :math:`(N_{nodes},)`

        Returns
        -------
        torch.Tensor
            Output tensor.
            Shape: :math:`(N_{nodes},d_e)`
        """
        x = self.concat(nodes, u[batch])

        heads = torch.stack([h(nodes, edges, edge_index, u, batch) for h in self.heads])
        gates = torch.stack([g(x).squeeze(-1) for g in self.gates])

        return torch.einsum("ijk, ij -> jk", heads, gates)


class EdgeAggregateGlobal(torch.nn.Module):
    def __init__(
        self,
        d_v: int,
        d_e: int,
        d_u: int,
        d_attn: int,
        num_heads: Optional[int] = 1,
    ) -> None:
        """Aggregates graph edges using multihead attention.

        Parameters
        ----------
        d_v
            Node feature length.
        d_e
            Edge feature length.
        d_u
            Global feature length.
        d_attn
            Attention vector length.
        num_heads
            Number of attention heads.
        """
        super().__init__()
        self.concat = luz.Concatenate(dim=1)
        self.heads = torch.nn.ModuleList()
        self.gates = torch.nn.ModuleList()

        for _ in range(num_heads):
            h = EdgeAggregateGlobalHead(d_v, d_e, d_u, d_attn)
            g = Dense(d_u, 1, activation=torch.nn.LeakyReLU())

            self.heads.append(h)
            self.gates.append(g)

    def forward(
        self,
        nodes: torch.Tensor,
        edges: torch.Tensor,
        edge_index: torch.Tensor,
        u: torch.Tensor,
        batch: torch.Tensor,
    ) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        nodes
            Node features.
            Shape: :math:`(N_{nodes},d_v)`
        edge_index
            Edge index tensor.
            Shape: :math:`(2,N_{edges})`
        edges
            Edge features, by default None.
            Shape: :math:`(N_{edges},d_e)`
        u
            Global features, by default None.
            Shape: :math:`(N_{batch},d_u)`
        batch
            Nodewise batch tensor, by default None.
            Shape: :math:`(N_{nodes},)`

        Returns
        -------
        torch.Tensor
            Output tensor.
            Shape: :math:`(N_{batch},d_e)`
        """
        heads = torch.stack([h(nodes, edges, edge_index, u, batch) for h in self.heads])
        gates = torch.stack([g(u).squeeze(-1) for g in self.gates])

        return torch.einsum("ijk, ij -> jk", heads, gates)


class NodeAggregate(torch.nn.Module):
    def __init__(
        self,
        d_v: int,
        d_u: int,
        num_heads: Optional[int] = 1,
    ) -> None:
        """Aggregates graph edges using multihead attention.

        Parameters
        ----------
        d_v
            Node feature length.
        d_u
            Global feature length.
        d_attn
            Attention vector length.
        num_heads
            Number of attention heads.
        """
        super().__init__()
        self.concat = luz.Concatenate(dim=1)
        self.heads = torch.nn.ModuleList()
        self.gates = torch.nn.ModuleList()

        for _ in range(num_heads):
            h = GraphConvAttention(d_v, activation=torch.nn.LeakyReLU())
            g = Dense(d_u, 1, activation=torch.nn.LeakyReLU())

            self.heads.append(h)
            self.gates.append(g)

    def forward(
        self,
        nodes: torch.Tensor,
        edges: torch.Tensor,
        edge_index: torch.Tensor,
        u: torch.Tensor,
        batch: torch.Tensor,
    ) -> torch.Tensor:
        """Compute forward pass.

        Parameters
        ----------
        nodes
            Node features.
            Shape: :math:`(N_{nodes},d_v)`
        edge_index
            Edge index tensor.
            Shape: :math:`(2,N_{edges})`
        edges
            Edge features, by default None.
            Shape: :math:`(N_{edges},d_e)`
        u
            Global features, by default None.
            Shape: :math:`(N_{batch},d_u)`
        batch
            Nodewise batch tensor, by default None.
            Shape: :math:`(N_{nodes},)`

        Returns
        -------
        torch.Tensor
            Output tensor.
            Shape: :math:`(N_{batch},d_v)`
        """
        heads = torch.stack([h(nodes, edge_index, batch) for h in self.heads])
        gates = torch.stack([g(u).squeeze(-1) for g in self.gates])

        return torch.einsum("ijk, ij -> jk", heads @ nodes, gates)



from __future__ import annotations
from typing import Any, Callable, Iterable, Optional, Tuple, Union

import itertools
import luz
import matplotlib.pyplot as plt
import numpy as np
import pathlib
import torch

__all__ = [
    "default_collate",
    "graph_collate",
    "Data",
    "Dataset",
    # "ChainDataset",
    "ConcatDataset",
    # "IterableDataset",
    "FolderDataset",
    "Subset",
    "TensorDataset",
    "UnpackDataset",
    "WrapperDataset",
]

Device = Union[str, torch.device]


def default_collate(batch: Iterable[luz.Data]) -> luz.Data:
    """Collate multiple Data objects.

    Parameters
    ----------
    batch : Iterable[luz.Data]
        Data objects to be collated

    Returns
    -------
    luz.Data
        Collated Data object
    """
    kw = {
        k: torch.stack([torch.as_tensor(sample[k]) for sample in batch], dim=0)
        for k in batch[0].keys
    }

    return Data(**kw)


def graph_collate(batch: Iterable[luz.Data]) -> luz.Data:
    """Collate multiple Data objects containing graph data.

    Parameters
    ----------
    batch : Iterable[luz.Data]
        Data objects to be collated

    Returns
    -------
    luz.Data
        Collated Data object
    """
    node_counts = [sample.x.shape[0] for sample in batch]
    edge_index_offsets = np.roll(np.cumsum(node_counts), shift=1)
    edge_index_offsets[0] = 0

    kw = {}

    for k in batch[0].keys:
        if k in ("x", "edges"):  # , "y"):
            kw[k] = torch.cat([torch.as_tensor(sample[k]) for sample in batch], dim=0)
        elif k == "edge_index":
            kw[k] = torch.cat(
                [
                    torch.as_tensor(sample[k] + offset, dtype=torch.long)
                    for sample, offset in zip(batch, edge_index_offsets)
                ],
                dim=1,
            )
        else:
            kw[k] = torch.stack([torch.as_tensor(sample[k]) for sample in batch], dim=0)

    kw["batch"] = torch.cat(
        [torch.full((nc,), i, dtype=torch.long) for i, nc in enumerate(node_counts)]
    )

    return Data(**kw)


class Data:
    def __init__(
        self, x: torch.Tensor, y: torch.Tensor = None, **kwargs: torch.Tensor
    ) -> None:
        """Data containing one or more tensors.

        Parameters
        ----------
        x
            Primary data tensor.
        y
            Label tensor, by default None.
        **kwargs
            Additional data tensors.
        """
        self.d = {"x": x, "y": y, **kwargs}

        for k, v in dict(x=x, y=y, **kwargs).items():
            setattr(self, k, v)

    @property
    def keys(self) -> None:
        return self.d.keys()

    def __getitem__(self, k: str) -> torch.Tensor:
        return self.d[k]

    def to(self, device: Device) -> luz.Data:
        """Migrate tensors to device. Modifies Data object in-place.

        Parameters
        ----------
        device
            Target device.

        Returns
        -------
        luz.Data
            Migrated Data.
        """
        for k, v in self.d.items():
            if torch.is_tensor(v):
                self.d[k] = v.to(device)

            setattr(self, k, self.d[k])

        return self

    def __repr__(self) -> str:
        kw_str = ",".join(f"{k}={v}" for k, v in self.d.items())
        return f"Data({kw_str})"


class BaseDataset:
    def _collate(self, batch: Iterable[luz.Data]) -> luz.Data:
        return default_collate(batch)

    def use_collate(
        self, collate: Callable[[Iterable[luz.Data]], luz.Data]
    ) -> luz._BaseDataset:
        self._collate = collate
        return self

    def loader(
        self,
        batch_size: Optional[int] = 1,
        shuffle: Optional[bool] = True,
        num_workers: Optional[int] = 1,
        pin_memory: Optional[bool] = True,
    ) -> torch.utils.data.Dataloader:
        """Generate Dataloader.

        Parameters
        ----------
        batch_size
            Batch size.
            By default 1.
        shuffle
            If True, shuffle dataset.
            By default True.
        num_workers
            Number of workers.
            By default 1.
        pin_memory
            If True, put fetched tensors in pinned memory.
            By default True.


        Returns
        -------
        torch.utils.data.Dataloader
            Generated Dataloader.
        """

        return torch.utils.data.DataLoader(
            dataset=self,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=pin_memory,
            collate_fn=self._collate,
        )

    def __add__(self, other: luz.Dataset) -> luz.ConcatDataset:
        return ConcatDataset([self, other])

    def subset(self, indices: Iterable[int]) -> luz.Subset:
        """Generate subset of this Dataset.

        Parameters
        ----------
        indices
            Indices used to select elements of the subset.

        Returns
        -------
        luz.Subset
            Generated subset.
        """
        return Subset(dataset=self, indices=indices).use_collate(self._collate)

    def split(
        self, lengths: Iterable[int], shuffle: Optional[int] = True
    ) -> Tuple[BaseDataset]:
        """Split Dataset into multiple subsets of given lengths.

        Parameters
        ----------
        lengths
            Lengths of subsets.
        shuffle
            If True, shuffle before splitting; by default True.

        Returns
        -------
        Tuple[BaseDataset]
            Generated subsets.
        """
        # adapted from
        # https://pytorch.org/docs/stable/_modules/torch/utils/data/dataset.html#random_split
        if shuffle:
            indices = torch.randperm(sum(lengths)).tolist()
        else:
            indices = torch.arange(sum(lengths)).tolist()
        return tuple(
            self.subset(indices=indices[offset - l : offset]).use_collate(self._collate)
            for offset, l in zip(itertools.accumulate(lengths), lengths)
        )

    def plot_histogram(
        self,
        key: str,
        dim: int,
        num_bins: Optional[int] = 100,
        data_min: Optional[float] = None,
        data_max: Optional[float] = None,
        filepath: Optional[Union[str, pathlib.Path]] = None,
    ) -> None:
        if data_min is None:
            data_min = self.min(key, dim)

        if data_max is None:
            data_max = self.max(key, dim)

        for x in self:
            histc = torch.zeros_like(
                torch.histc(x[key][dim], bins=num_bins, min=data_min, max=data_max)
            )
            break

        for i in range(len(self)):
            histc += torch.histc(
                self[i][key][dim],
                bins=num_bins,
                min=data_min,
                max=data_max,
            )

        bins = np.linspace(data_min, data_max, num_bins + 1)

        plt.hist(bins[:-1], bins, weights=histc.numpy())
        if filepath is not None:
            plt.savefig(luz.expand_path(filepath))
        plt.show()


class Dataset(torch.utils.data.Dataset, BaseDataset):
    def __init__(self, data: Iterable[luz.Data]) -> None:
        """Object containing points from a domain, possibly with labels.

        Parameters
        ----------
        data
            Iterable of Data objects comprising the dataset.
        """
        self.data = tuple(data)
        self.len = len(self.data)

    def __getitem__(self, index: int) -> luz.Data:
        return self.data[index]

    def __len__(self) -> int:
        return self.len


class TensorDataset(torch.utils.data.Dataset, BaseDataset):
    def __init__(self, **tensors: torch.Tensor) -> None:
        ks = tensors.keys()
        self.data = tuple(
            Data(**dict(zip(ks, torch.atleast_1d(z)))) for z in zip(*tensors.values())
        )
        self.len = len(self.data)

    def __getitem__(self, index: int) -> luz.Data:
        return self.data[index]

    def __len__(self) -> int:
        return self.len


class ConcatDataset(BaseDataset, torch.utils.data.ConcatDataset):
    pass


class Subset(BaseDataset, torch.utils.data.Subset):
    pass


class FolderDataset(BaseDataset, torch.utils.data.Dataset):
    def __init__(self, root: str) -> None:
        """Dataset which reads data from disk.

        Parameters
        ----------
        root
            Root directory containing data stored in .pt files.
        """
        self.root = luz.expand_path(root)

    def __getitem__(self, index: int) -> luz.Data:
        return torch.load(pathlib.Path(self.root, f"{index}.pt"))

    def __len__(self) -> int:
        return len(tuple(pathlib.Path(self.root).glob("[0-9]*.pt")))


class UnpackDataset(BaseDataset, torch.utils.data.Dataset):
    def __init__(self, keys: Iterable[str], dataset: torch.utils.data.Dataset) -> None:
        self.keys = keys
        self.dataset = dataset

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, index: int) -> luz.Data:
        return Data(**dict(zip(self.keys, self.dataset[index])))


class WrapperDataset(BaseDataset, torch.utils.data.Dataset):
    def __init__(self, **datasets: torch.utils.data.Dataset) -> None:
        self.datasets = datasets

    def __len__(self) -> int:
        d, *_ = self.datasets.values()
        return len(d)

    def __getitem__(self, index: int) -> Any:
        return Data(**{k: d[index] for k, d in self.datasets.items()})


class ChainDataset(BaseDataset, torch.utils.data.ChainDataset):
    pass



from __future__ import annotations
from typing import Any, Iterable, Optional, Union

import luz

# import scipy.optimize
import functools
import torch

__all__ = [
    "Argmax",
    "Compose",
    "Identity",
    "Lookup",
    "NanToNum",
    "NormalizePerTensor",
    "PowerSeries",
    "Reshape",
    "Scale",
    "Squeeze",
    "TensorTransform",
    "Transform",
    "Transpose",
    "Unsqueeze",
    "YeoJohnson",
    "ZeroMeanPerTensor",
]


class Transform(torch.nn.Module):
    def __init__(self, **transforms: TensorTransform) -> None:
        super().__init__()
        self.transforms = torch.nn.ModuleDict(transforms)

    def forward(self, data: luz.Data) -> luz.Data:
        """Transform data.

        Parameters
        ----------
        data
            Input data.

        Returns
        -------
        luz.Data
            Output data.
        """
        kw = {
            k: self.transforms[k](data[k]) if k in self.transforms else data[k]
            for k in data.keys
        }
        return luz.Data(**kw)

    def __mul__(self, other):
        kw = {}

        for k in self.transforms:
            if k in other.transforms:
                kw[k] = luz.Compose(self.transforms[k], other.transforms[k])
            else:
                kw[k] = self.transforms[k]

        for k in other.transforms:
            if k not in self.transforms:
                kw[k] = other.transforms[k]

        return Transform(**kw)

    def attach(self, runner: luz.Runner):

        runner.run_batch = self.dec(runner.run_batch)

    def dec(self, f):
        @functools.wraps(f)
        def g(state, batch):
            return f(state, self.forward(batch))

        return g


class TensorTransform(torch.nn.Module):
    def inverse(self) -> TensorTransform:
        raise NotImplementedError


# INVERTIBLE


class Compose(TensorTransform):
    def __init__(self, *transforms: Iterable[TensorTransform]) -> None:
        super().__init__()
        self.transforms = torch.nn.ModuleList(transforms)

    def forward(self, x: Any) -> Any:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        for transform in self.transforms:
            x = transform(x)
        return x

    def inverse(self) -> TensorTransform:
        inverse_transforms = [t.inverse() for t in self.transforms][::-1]
        return Compose(*inverse_transforms)


class Squeeze(TensorTransform):
    def __init__(self, dim: Optional[int]) -> None:
        """Squeeze tensor.

        Parameters
        ----------
        dim
            Dimension to be squeezed.
        """
        super().__init__()
        self.dim = dim

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Squeezed output tensor.
        """
        return x.squeeze(dim=self.dim)

    def inverse(self) -> TensorTransform:
        return Unsqueeze(self.dim)


class Unsqueeze(TensorTransform):
    def __init__(self, dim: Optional[int]) -> None:
        """Unsqueeze tensor.

        Parameters
        ----------
        dim
            Dimension to be unsqueezed.
        """
        super().__init__()
        self.dim = dim

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Unsqueezed output tensor.
        """
        return x.unsqueeze(dim=self.dim)

    def inverse(self) -> TensorTransform:
        return Squeeze(self.dim)


class Identity(TensorTransform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor, same as the input tensor.
        """
        return x

    def inverse(self) -> TensorTransform:
        return Identity()


class YeoJohnson(TensorTransform):
    def __init__(self):
        super().__init__()
        self.lmbda = None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = torch.zeros_like(x)
        lmbda = self.lmbda.expand(*x.shape)
        pos = x >= 0

        lmzro = lmbda[pos] == 0

        out.ravel().scatter_(
            0, pos.ravel().nonzero().squeeze(-1)[lmzro], torch.log1p(x[pos][lmzro])
        )
        out.ravel().scatter_(
            0,
            pos.ravel().nonzero().squeeze(-1)[~lmzro],
            ((x[pos][~lmzro] + 1) ** lmbda[pos][~lmzro] - 1) / lmbda[pos][~lmzro],
        )

        lmtwo = lmbda[~pos] == 2

        out.ravel().scatter_(
            0,
            (~pos).ravel().nonzero().squeeze(-1)[lmtwo],
            -torch.log1p(-x[~pos][lmtwo]),
        )
        out.ravel().scatter_(
            0,
            (~pos).ravel().nonzero().squeeze(-1)[~lmtwo],
            -((-x[~pos][~lmtwo] + 1) ** (2 - lmbda[~pos][~lmtwo]) - 1)
            / (2 - lmbda[~pos][~lmtwo]),
        )

        return out


class Scale(TensorTransform):
    def __init__(self, shift: torch.Tensor, scale: torch.Tensor) -> None:
        super().__init__()
        self.shift = shift
        self.scale = scale

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Shifted and scaled output tensor.
        """
        return (x - self.shift) / self.scale

    def inverse(self) -> TensorTransform:
        return Scale(-self.shift / self.scale, 1 / self.scale)


class Transpose(TensorTransform):
    def __init__(self, dim0: int, dim1: int) -> None:
        super().__init__()
        self.dim0 = dim0
        self.dim1 = dim1

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Transposed output tensor.
        """
        return torch.transpose(x, self.dim0, self.dim1)

    def inverse(self) -> TensorTransform:
        return Transpose(self.dim0, self.dim1)


# NON-INVERTIBLE


class Argmax(TensorTransform):
    def __init__(
        self, dim: Optional[int] = None, keepdim: Optional[bool] = False
    ) -> None:
        super().__init__()
        self.dim = dim
        self.keepdim = keepdim

    def forward(self, x: torch.Tensor) -> torch.LongTensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        return x.argmax(dim=self.dim, keepdim=self.keepdim)


class Lookup(TensorTransform):
    def __init__(self, lookup_dict) -> None:
        super().__init__()
        self.lookup_dict = lookup_dict

    def forward(self, x) -> Any:
        """Transform tensor.

        Parameters
        ----------
        x
            Input.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        return self.lookup_dict[x]


class NanToNum(TensorTransform):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.args = args
        self.kwargs = kwargs

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nan_to_num(x, *self.args, **self.kwargs)


class NormalizePerTensor(TensorTransform):
    def __init__(self, p: Union[int, str], *args, **kwargs) -> None:
        super().__init__()
        self.p = p
        self.args = args
        self.kwargs = kwargs

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Normalized output tensor.
        """
        return x / torch.linalg.norm(x, self.p, *self.args, **self.kwargs)


class PowerSeries(TensorTransform):
    def __init__(self, degree: int) -> None:
        super().__init__()
        self.degree = degree

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        return torch.vander(x, self.degree + 1)


class Reshape(TensorTransform):
    def __init__(self, out_shape: Iterable[int]) -> None:
        """Reshape tensor.

        Parameters
        ----------
        out_shape
            Desired output shape.
        """
        super().__init__()
        self.shape = tuple(out_shape)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Reshaped output tensor.
        """
        return x.view(self.shape)


class ZeroMeanPerTensor(TensorTransform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Transform tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        return x - torch.mean(x)



from __future__ import annotations
from typing import Callable, Optional, Union

import luz
import torch

__all__ = ["Learner"]

Device = Union[str, torch.device]
Criterion = Callable[[torch.Tensor, torch.Tensor], torch.Tensor]


class Learner:
    def __init__(self) -> None:
        self.hparams = None
        self.preprocessor = None
        self.trainer = None
        self.validator = None
        self.tester = None

    def model(self) -> torch.nn.Module:
        """Get model."""
        raise NotImplementedError

    def optimizer(self, model: torch.nn.Module) -> torch.optim.Optimizer:
        """Get optimizer."""
        raise NotImplementedError

    def criterion(self) -> Criterion:
        """Get loss function."""
        raise NotImplementedError

    def loader(self, dataset: luz.Dataset, stage: str) -> torch.utils.data.DataLoader:
        """Get dataloader."""
        return dataset.loader()

    def transform(self, state: luz.State) -> None:
        """Create and apply transform to runners.

        Parameters
        ----------
        state
            Preprocessor state.
        """
        pass

    def callbacks(self, runner: luz.Runner, stage: str) -> None:
        """Apply callbacks to runners."""
        pass

    def learn(
        self,
        train_dataset: luz.Dataset,
        val_dataset: Optional[luz.Dataset] = None,
        test_dataset: Optional[luz.Dataset] = None,
        device: Optional[Device] = "cpu",
    ) -> torch.nn.Module:
        """Learn a model based on a given dataset.

        Parameters
        ----------
        train_dataset
            Training dataset used to learn a model.
        val_dataset
            Validation dataset, by default None.
        test_dataset
            Test dataset, by default None.
        device
            Device to use for learning, by default "cpu".

        Returns
        -------
        torch.nn.Module
            Learned model.
        """
        model = self.model().to(device)
        optimizer = self.optimizer(model)
        criterion = self.criterion()

        self._runners(model, train_dataset, val_dataset, test_dataset)

        self.trainer.RUNNER_STARTED.attach(
            luz.UpdateState(optimizer=optimizer, criterion=criterion)
        )

        if self.preprocessor is not None:
            self.preprocessor.RUNNER_STARTED.attach(
                luz.UpdateState(criterion=criterion)
            )
            self.preprocessor.RUNNER_ENDED.attach(self._attach_transform)
            self.trainer.RUNNER_STARTED.attach(luz.Run(self.preprocessor, device))
            self.callbacks(self.preprocessor, "preprocess")

        if val_dataset is not None:
            self.validator.RUNNER_STARTED.attach(luz.UpdateState(criterion=criterion))
            self.trainer.EPOCH_ENDED.attach(luz.Run(self.validator, device))
            self.callbacks(self.validator, "validate")

        if test_dataset is not None:
            self.tester.RUNNER_STARTED.attach(luz.UpdateState(criterion=criterion))
            self.trainer.RUNNER_ENDED.attach(luz.Run(self.tester, device))
            self.callbacks(self.tester, "test")

        self.callbacks(self.trainer, "train")

        self.trainer.run(device=device)

        if test_dataset is not None:
            return self.trainer.state.model, self.tester.state.metrics["loss"]
        return self.trainer.state.model

    def _runners(
        self,
        model: torch.nn.Module,
        train_dataset: luz.Dataset,
        val_dataset: Optional[luz.Dataset] = None,
        test_dataset: Optional[luz.Dataset] = None,
    ) -> None:
        self.trainer = self.runner(model, train_dataset, "train")
        self.preprocessor = self.runner(model, train_dataset, "preprocess")
        if val_dataset is not None:
            self.validator = self.runner(model, val_dataset, "validate")
        if test_dataset is not None:
            self.tester = self.runner(model, test_dataset, "test")

    def _attach_transform(self, state: luz.State) -> None:
        transform = self.transform(state)
        for runner in [self.trainer, self.validator, self.tester]:
            if runner is not None:
                runner.state.update(transform=transform)



from __future__ import annotations
from typing import Any, Callable, Optional, Union

from abc import ABC, abstractmethod
import datetime
import luz
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pathlib
import torch

__all__ = [
    "Accuracy",
    "CalibrationPlot",
    "DurbinWatson",
    "FBeta",
    "HistogramResiduals",
    "LearningCurvePlot",
    "Loss",
    "Max",
    "MeanStd",
    "Metric",
    "Min",
    "RegressionPlot",
    "ResidualPlot",
    "TimeEpochs",
]

Path = Optional[Union[str, pathlib.Path]]


class Metric(ABC):
    @abstractmethod
    def reset(self) -> None:
        """Reset metric state."""
        pass

    @abstractmethod
    def update(self, **kwargs: Any) -> None:
        """Update metric state."""
        pass

    @abstractmethod
    def compute(self) -> Any:
        """Compute metric."""
        pass

    def reset_handler(self, state: luz.State) -> None:
        self.reset()

    def update_handler(self, state: luz.State) -> None:
        self.update(**state.__dict__)

    def get_compute_handler(self, name: str) -> Callable[[luz.State], None]:
        def compute_handler(state: luz.State) -> None:
            state.metrics[name] = self.compute()

        return compute_handler

    def attach_events(
        self, runner: luz.Runner
    ) -> tuple[luz.Event, luz.Event, luz.Event]:
        """Get events to attach metric steps.

        Parameters
        ----------
        luz.Runner
            Runner to which metric will be attached.

        Returns
        -------
        tuple[luz.Event, luz.Event, luz.Event]
            Runner events to which reset, update, and compute steps are attached.
        """
        return runner.EPOCH_STARTED, runner.BATCH_ENDED, runner.EPOCH_ENDED

    def attach(
        self,
        runner: luz.Runner,
        name: str,
    ) -> None:
        reset_event, update_event, compute_event = self.attach_events(runner)
        reset_event.attach(self.reset_handler)
        update_event.attach(self.update_handler)
        compute_event.attach(self.get_compute_handler(name))


class Accuracy(Metric):
    def reset(self) -> None:
        """Reset metric state."""
        self.correct = 0
        self.total = 0

    def update(self, output: torch.Tensor, target: torch.Tensor, **kwargs: Any) -> None:
        """Update metric state.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,C)`
        target
            Target tensor. One-hot encoded.
            Shape: :math:`(N,C)`
        """
        predicted = torch.argmax(torch.softmax(output, dim=1), dim=1)
        correct = torch.argmax(target, dim=1)

        self.correct += (predicted == correct).sum().item()
        self.total += target.size(0)

    def compute(self) -> float:
        """Compute metric."""
        return self.correct / self.total


class CalibrationPlot(Metric):
    def __init__(
        self, filepath: Optional[Path] = None, rasterized: Optional[bool] = False
    ) -> None:
        """Plot actual labels vs. predicted labels.

        Parameters
        ----------
        filepath
            Path to save plot if not None, by default None.
        rasterized
            Rasterize data points if True, by default False.
        """
        self.filepath = filepath
        self.rasterized = rasterized

    def reset(self) -> None:
        """Reset metric state."""
        self.fig, self.ax = plt.subplots()

    def update(self, output: torch.Tensor, target: torch.Tensor, **kwargs: Any) -> None:
        """Update metric state.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,)`
        target
            Target tensor.
            Shape: :math:`(N,)`
        """
        x = output.cpu().detach().numpy().reshape(-1)
        y = target.cpu().detach().numpy().reshape(-1)
        self.ax.scatter(x, y, color="black", rasterized=self.rasterized)

    def compute(self) -> plt.Figure:
        """Compute metric."""
        self.ax.relim()
        self.ax.autoscale_view()

        line = matplotlib.lines.Line2D([0, 1], [0, 1], color="red")
        line.set_transform(self.ax.transAxes)
        self.ax.add_line(line)

        self.ax.set_xlabel("Predicted")
        self.ax.set_ylabel("Actual")
        self.ax.set_title("Actual vs. predicted")

        if self.filepath is not None:
            self.fig.savefig(luz.expand_path(self.filepath))

        return self.fig


class DurbinWatson(Metric):
    def reset(self) -> None:
        """Reset metric state."""
        self.num = 0.0
        self.denom = 0.0
        self.last_residual = 0.0

    def update(self, output: torch.Tensor, target: torch.Tensor, **kwargs: Any) -> None:
        """Update metric state."""
        residual = target.cpu().detach() - output.cpu().detach()
        diffs = torch.diff(residual, dim=0)

        self.num += (diffs ** 2).sum(dim=0) + (residual[0] - self.last_residual) ** 2
        self.denom += (residual ** 2).sum(dim=0)

        self.last_residual = residual[-1]

    def compute(self) -> float:
        """Compute metric."""
        return self.num / self.denom


class FBeta(Metric):
    def __init__(self, beta: float) -> None:
        self.beta = beta

    def reset(self, **kwargs: Any) -> None:
        """Reset metric state."""
        self.true_positive = 0
        self.predicted_positive = 0
        self.actual_positive = 0

    def update(self, output: torch.Tensor, target: torch.Tensor, **kwargs: Any) -> None:
        """Update metric state.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,2)`
        target
            Target tensor. One-hot encoded.
            Shape: :math:`(N,2)`
        """
        predicted = torch.argmax(torch.softmax(output, dim=1), dim=1)
        correct = torch.argmax(target, dim=1)

        self.true_positive += correct[predicted.nonzero(as_tuple=False)].sum().item()
        self.predicted_positive += predicted.sum().item()
        self.actual_positive += correct.sum().item()

    def compute(self) -> float:
        """Compute metric."""
        try:
            precision = self.true_positive / self.predicted_positive
            recall = self.true_positive / self.actual_positive
            return (
                (1 + self.beta ** 2)
                * precision
                * recall
                / (precision * (self.beta ** 2) + recall)
            )
        except ZeroDivisionError:
            return 1.0


class HistogramData(Metric):
    def __init__(
        self,
        filepath: Optional[Path] = None,
        num_bins: Optional[int] = 100,
        rasterized: Optional[bool] = False,
    ) -> None:
        """Plot actual labels vs. predicted labels.

        Parameters
        ----------
        filepath
            Path to save plot if not None, by default None.
        num_bins
            Number of histogram bins, by default 100.
        rasterized
            Rasterize data points if True, by default False.
        """
        self.filepath = filepath
        self.num_bins = num_bins
        self.rasterized = rasterized

    def reset(self) -> None:
        """Reset metric state."""
        self.fig, self.ax = plt.subplots()
        self.bins = np.linspace(self.data_min, self.data_max, self.num_bins + 1)
        self.histc = None

    def update(self, x: torch.Tensor, **kwargs: Any) -> None:
        """Update metric state.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,)`
        target
            Target tensor.
            Shape: :math:`(N,)`
        """
        if self.histc is None:
            self.histc = torch.histc(
                x, bins=self.num_bins, min=self.data_min, max=self.data_max
            )
        else:
            self.histc += torch.histc(
                x, bins=self.num_bins, min=self.data_min, max=self.data_max
            )

    def compute(self) -> plt.Figure:
        """Compute metric."""
        plt.hist(self.bins[:-1], self.bins, weights=self.histc.numpy())
        if self.filepath is not None:
            plt.savefig(luz.expand_path(self.filepath))

        self.ax.relim()
        self.ax.autoscale_view()

        line = matplotlib.lines.Line2D([0, 1], [0, 1], color="red")
        line.set_transform(self.ax.transAxes)
        self.ax.add_line(line)

        self.ax.set_xlabel("Predicted")
        self.ax.set_ylabel("Actual")
        self.ax.set_title("Actual vs. predicted")

        if self.filepath is not None:
            self.fig.savefig(luz.expand_path(self.filepath))

        return self.fig


class HistogramResiduals(Metric):
    def __init__(
        self,
        filepath: Optional[Path] = None,
        num_bins: Optional[int] = 100,
    ) -> None:
        """Histogram residuals.

        Parameters
        ----------
        filepath
            Path to save plot if not None, by default None.
        num_bins
            Number of histogram bins, by default 100.
        """
        self.filepath = filepath
        self.num_bins = num_bins

    def reset(self) -> None:
        """Reset metric state."""
        self.fig, self.ax = plt.subplots()
        self.residuals = []

    def update(self, output: torch.Tensor, target: torch.Tensor, **kwargs: Any) -> None:
        """Update metric state.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,C)`
        target
            Target tensor.
            Shape: :math:`(N,C)`
        """
        r = (output - target).cpu().detach().reshape(-1).numpy()
        self.residuals.extend(r)

    def compute(self) -> plt.Figure:
        """Compute metric."""
        self.ax.hist(self.residuals, bins=self.num_bins, edgecolor="black")

        self.ax.axvline(x=0, color="r", linestyle="--")

        self.ax.set_xlabel("Residual")
        self.ax.set_ylabel("Count")
        self.ax.set_title("Residuals")

        if self.filepath is not None:
            self.fig.savefig(luz.expand_path(self.filepath))

        return self.fig


class LearningCurvePlot(Metric):
    def __init__(
        self,
        # loss_func,
        filepath: Optional[Path] = None,
        reduction: Optional[str] = "mean",
    ) -> None:
        # self.loss_func = loss_func
        self.filepath = filepath
        self.reduction = reduction
        self.epoch_losses = []

    # def attach_events(
    #     self, runner
    #     ) -> tuple[luz.Event, luz.Event, luz.Event]:
    #     """Get events to attach metric steps.

    #     Parameters
    #     ----------
    #     luz.Runner
    #         Runner to which metric will be attached.

    #     Returns
    #     -------
    #     tuple[luz.Event, luz.Event, luz.Event]
    #         Runner events to which reset, update, and compute steps are attached.
    #     """
    #     return runner.RUNNER_STARTED, runner.EPOCH_ENDED, runner.RUNNER_ENDED

    def reset(self) -> None:
        """Reset metric state."""
        self.fig, self.ax = plt.subplots()
        [self.history] = self.ax.plot([], [])

    def update(self, ind: int, epoch: int, loss: torch.Tensor, **kwargs: Any) -> None:
        """Update metric state.

        Parameters
        ----------
        epoch
            Epoch number.
        loss
            Loss tensor.
            Shape: :math:`(1,)`
        """
        delta = loss.item() - self.mean_loss
        self.mean_loss += delta / (ind + 1)

        fit_xdata, fit_ydata = self.history.get_data()
        xdata = np.append(fit_xdata, epoch + 1)
        loss = self.loss_func()
        ydata = np.append(fit_ydata, loss.item())
        self.history.set(xdata=xdata, ydata=ydata)

    def compute(self) -> plt.Figure:
        """Compute metric."""
        fit_xdata, fit_ydata = self.history.get_data()
        argsort = fit_xdata.argsort()
        self.history.set(xdata=fit_xdata[argsort], ydata=fit_ydata[argsort])

        self.ax.relim()
        self.ax.autoscale_view()

        self.ax.set_xlabel("Epoch")
        self.ax.set_ylabel("Loss")
        self.ax.set_title("Loss history")

        self.fig.tight_layout()

        if self.filepath is not None:
            self.fig.savefig(luz.expand_path(self.filepath))

        return self.fig

    # def __call__(
    #     self,
    #     state: luz.State,
    # ) -> None:
    #     history = state.history
    #     loader = state.loader

    #     if self.reduction == "mean":
    #         # divide each epoch loss (which is the sum of batch averages)
    #         # by the number of batches to estimate average epoch loss
    #         y1 = np.array(history) / len(loader)

    #     #(line1,) = ax1.plot(x, y1, color="tab:blue")

    #     if len(history) > 0:
    #         ax2 = ax1.twinx()

    #         if self.reduction == "mean":
    #             # divide each epoch loss (which is the sum of batch averages)
    #             # by the number of batches to estimate average epoch loss
    #             y2 = np.array(history) / len(loader)

    #         (line2,) = ax2.plot(x, y2, color="tab:orange")

    #         lines = (line1, line2)
    #         labels = (
    #             f"Training loss (min: {min(history)})",
    #             f"Validation loss (min: {min(history)})",
    #         )
    #     else:
    #         lines = (line1,)
    #         labels = (f"Training loss (min: {min(history)})",)

    #     plt.title("Loss history")
    #     plt.legend(lines, labels)

    #     fig.tight_layout()

    #     if self.save_filepath is not None:
    #         plt.savefig(self.save_filepath)

    #     if self.show_plot:
    #         plt.show()
    #     else:
    #         plt.close(fig)


class Loss(Metric):
    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        """Reset metric state."""
        self.mean_loss = 0

    def update(
        self,
        loss: torch.Tensor,
        ind: int,
        **kwargs: Any,
    ) -> None:
        """Update metric state.

        Parameters
        ----------
        loss
            Last computed loss.
        ind
            Data index.
        **kwargs
            Superfluous kwargs.
        """
        delta = loss.item() - self.mean_loss
        self.mean_loss += delta / (ind + 1)

    def compute(self) -> float:
        """Compute metric."""
        return self.mean_loss


class Max(Metric):
    def __init__(self, key: Optional[str] = "x", batch_dim: Optional[int] = 0) -> None:
        """Compute maximum of data.

        Parameters
        ----------
        key
            Data key, by default "x".
        batch_dim
            Batch dimension, by default 0.
        """
        self.key = key
        self.batch_dim = batch_dim

    def reset(self) -> None:
        """Reset metric state."""
        self.max = torch.Tensor([float("-inf")])

    def update(self, data: luz.Data, **kwargs: Any) -> None:
        """Update metric state."""
        a, _ = torch.max(self.max, dim=self.batch_dim)
        b, _ = torch.max(data[self.key], dim=self.batch_dim)
        self.max = torch.max(a, b)

    def compute(self) -> torch.Tensor:
        """Compute metric."""
        return self.max


class MeanStd(Metric):
    def __init__(self, key: Optional[str] = "x", batch_dim: Optional[int] = 0) -> None:
        """Compute mean and standard deviation of data.

        Parameters
        ----------
        key
            Data key, by default "x".
        batch_dim
            Batch dimension, by default 0.
        """
        self.key = key
        self.batch_dim = batch_dim

    def reset(self) -> None:
        """Reset metric state."""
        self.mean = 0.0
        self.var = 0.0
        self.n = 0.0

    def update(self, data: luz.Data, **kwargs: Any) -> None:
        """Update metric state."""
        x = data[self.key].cpu().detach()
        self.n = x.size(self.batch_dim)
        delta = x.cpu().detach() - self.mean
        self.mean += delta.sum(self.batch_dim) / self.n
        self.var += (delta * (x.cpu().detach() - self.mean)).sum(self.batch_dim)

    def compute(self) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute metric."""
        return self.mean, torch.sqrt(self.var / self.n)


class Min(Metric):
    def __init__(self, key: Optional[str] = "x", batch_dim: Optional[int] = 0) -> None:
        """Compute minimum of data.

        Parameters
        ----------
        key
            Data key, by default "x".
        batch_dim
            Batch dimension, by default 0.
        """
        self.key = key
        self.batch_dim = batch_dim

    def reset(self) -> None:
        """Reset metric state."""
        self.min = torch.Tensor([float("inf")])

    def update(self, data: luz.Data, **kwargs: Any) -> None:
        """Update metric state."""
        a, _ = torch.min(self.min, dim=self.batch_dim)
        b, _ = torch.min(data[self.key], dim=self.batch_dim)
        self.min = torch.min(a, b)

    def compute(self) -> torch.Tensor:
        """Compute metric."""
        return self.min


class RegressionPlot(Metric):
    def __init__(
        self, filepath: Optional[Path] = None, rasterized: Optional[bool] = False
    ) -> None:
        """Plot data and regression.

        Parameters
        ----------
        filepath
            Path to save plot if not None, by default None.
        rasterized
            Rasterize data points if True, by default False.
        """
        self.filepath = filepath
        self.rasterized = rasterized

    @property
    def name(self) -> str:
        """Metric name."""
        return "regression_plot"

    def reset(self) -> None:
        """Reset metric state."""
        self.fig, self.ax = plt.subplots()
        [self.fit] = self.ax.plot([], [])

    def update(
        self, x: torch.Tensor, output: torch.Tensor, target: torch.Tensor, **kwargs: Any
    ) -> None:
        """Update metric state.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,C)`
        target
            Target tensor.
            Shape: :math:`(N,C)`
        """
        x = x.cpu().detach().reshape(-1).numpy()
        y = target.cpu().detach().reshape(-1).numpy()
        y_fit = output.cpu().detach().reshape(-1).numpy()

        self.ax.scatter(x, y, color="black", label="Data", rasterized=self.rasterized)

        fit_xdata, fit_ydata = self.fit.get_data()
        xdata = np.append(fit_xdata, x)
        ydata = np.append(fit_ydata, y_fit)
        self.fit.set(xdata=xdata, ydata=ydata)

    def compute(self) -> plt.Figure:
        """Compute metric."""
        fit_xdata, fit_ydata = self.fit.get_data()
        argsort = fit_xdata.argsort()
        self.fit.set(xdata=fit_xdata[argsort], ydata=fit_ydata[argsort])

        self.ax.relim()
        self.ax.autoscale_view()

        self.ax.set_xlabel("Input")
        self.ax.set_ylabel("Output")
        self.ax.set_title("Data and fit function")

        if self.filepath is not None:
            self.fig.savefig(luz.expand_path(self.filepath))

        return self.fig


class ResidualPlot(Metric):
    def __init__(
        self, filepath: Optional[Path] = None, rasterized: Optional[bool] = False
    ) -> None:
        """Plot residuals.

        Parameters
        ----------
        filepath
            Path to save plot if not None, by default None.
        rasterized
            Rasterize data points if True, by default False.
        """
        self.filepath = filepath
        self.rasterized = rasterized

    def reset(self) -> None:
        """Reset metric state."""
        self.fig, self.ax = plt.subplots()

    def update(
        self, x: torch.Tensor, output: torch.Tensor, target: torch.Tensor, **kwargs: Any
    ) -> None:
        """Update metric state.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,C)`
        target
            Target tensor.
            Shape: :math:`(N,C)`
        """
        x = x.cpu().detach().reshape(-1).numpy()
        r = (output - target).cpu().detach().reshape(-1).numpy()

        self.ax.scatter(x, r, color="black", rasterized=self.rasterized)

    def compute(self) -> plt.Figure:
        """Compute metric."""
        self.ax.relim()
        self.ax.autoscale_view()

        self.ax.axhline(y=0, color="r", linestyle="--")

        self.ax.set_xlabel("Input")
        self.ax.set_ylabel("Residual")
        self.ax.set_title("Residuals")

        if self.filepath is not None:
            self.fig.savefig(luz.expand_path(self.filepath))

        return self.fig


class TimeEpochs(Metric):
    def attach_events(self, runner) -> tuple[luz.Event, luz.Event, luz.Event]:
        """Get events to attach metric steps.

        Parameters
        ----------
        luz.Runner
            Runner to which metric will be attached.

        Returns
        -------
        tuple[luz.Event, luz.Event, luz.Event]
            Runner events to which reset, update, and compute steps are attached.
        """
        return runner.EPOCH_STARTED, runner.EPOCH_ENDED, runner.EPOCH_ENDED

    def reset(self) -> None:
        """Reset metric state."""
        self.start_time = datetime.datetime.now()  # .replace(microsecond=0)
        self.end_time = None

    def update(self, **kwargs: Any) -> None:
        """Update metric state."""
        self.end_time = datetime.datetime.now()  # .replace(microsecond=0)

    def compute(self):
        """Compute metric."""
        return self.end_time - self.start_time


class YeoJohnsonNLL(Metric):
    def __init__(self, lmbda: torch.Tensor) -> None:
        """Negative log loss for Yeo-Johnson transform fitting.

        Parameters
        ----------
        lmbda
            Lambda tensor.
        """
        self.lmbda = lmbda

    def reset(self) -> None:
        """Reset metric state."""
        self.loglike = 0.0
        self.mean = 0.0
        self.variance = 0.0
        self.denom = 0.0

    def update(self, x: torch.Tensor, **kwargs: Any) -> None:
        """Update metric state."""
        self.denom += x.shape[0]

        x_trans = torch.stack([self.forward(_x) for _x in x])
        delta = x_trans - self.mean
        self.mean += delta.sum(0) / self.denom
        self.variance += (delta * (x_trans - self.mean)).sum(0)

        self.loglike += (torch.sign(x) * torch.log1p(torch.abs(x))).sum(0)

    def compute(self) -> torch.Tensor:
        """Compute metric."""
        nll = self.loglike
        nll *= self.lmbda - 1
        nll += -self.denom / 2 * torch.log(self.variance / self.denom)
        nll = -nll.numpy()

        return nll

        # for i in range(len(self.lmbda)):
        # scipy.optimize.brent(objective(i), brack=(-2.0, 2.0))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['luz']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.3.2,<4.0.0',
 'pyee>=9.0.3,<10.0.0',
 'scipy>=1.7.2,<2.0.0',
 'torch>=1.7.0,<2.0.0']

setup_kwargs = {
    'name': 'luz',
    'version': '10.2.1',
    'description': 'Lightweight framework for structuring arbitrary reproducible neural network learning procedures using PyTorch.',
    'long_description': '==========\nLuz Module\n==========\n\n.. image:: https://codecov.io/gh/kijanac/luz/branch/main/graph/badge.svg\n  :target: https://codecov.io/gh/kijanac/luz\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github\n\n**Lightweight framework for structuring arbitrary reproducible neural network learning procedures using PyTorch.**\n\nPyTorch code can easily become complex, unwieldy, and difficult to understand as a project develops. Luz aims to provide a common scaffold for PyTorch code in order to minimize boilerplate, maximize readability, and maintain the flexibility of PyTorch itself.\n\nThe basis of Luz is the Runner, an abstraction representing batch-wise processing of data over multiple epochs. Runner has predefined hooks to which code can be attached and a State which can be manipulated to define essentially arbitrary behavior. These hooks can be used to compose multiple Runners into a single algorithm, enabling dataset preprocessing, model testing, and other common tasks.\n\nTo further reduce boilerplate, the Learner abstraction is introduced as shorthand for the extremely common Preprocess-Train-Validate-Test algorithm. Simply inherit luz.Learner and define a handful of methods to completely customize your learning algorithm.\n\nTwo additional abstractions are provided for convenience: Scorers, which score (i.e. evaluate) a Learner according to some predefined procedure, and Tuners, which tune Learner hyperparameters. These abstractions provide a common interface which makes model selection a two-line process.\n\n---------------\nGetting Started\n---------------\n\nInstalling\n----------\nFrom `pip <https://pypi.org/project/luz/>`_:\n\n``pip install luz``\n\nFrom `conda <https://anaconda.org/kijana/luz>`_:\n\n``conda install -c conda-forge -c pytorch -c kijana luz``\n\nDocumentation\n-------------\nSee documentation `here <https://kijanac.github.io/luz/>`_.\n\nExamples\n--------\nSee example scripts in `Examples <https://github.com/kijanac/luz/tree/main/examples>`_.\n\n-------\nAuthors\n-------\nKi-Jana Carter\n\n-------\nLicense\n-------\nThis project is licensed under the MIT License - see the `LICENSE <https://github.com/kijanac/luz/blob/main/LICENSE>`_ file for details.\n\n------------\nContributing\n------------\nSee `CONTRIBUTING <https://github.com/kijanac/luz/blob/main/CONTRIBUTING.rst>`_.',
    'author': 'Ki-Jana Carter',
    'author_email': 'kijana@mit.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kijanac/luz',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



from __future__ import annotations

from abc import ABC, abstractmethod

__all__ = ["ConsoleLogger", "FileLogger", "Logger"]


class Logger(ABC):
    @abstractmethod
    def log(self, msg: str) -> None:
        pass


class ConsoleLogger(Logger):
    """Logs messages to console."""

    def log(self, msg: str) -> None:
        print(msg)


class FileLogger(Logger):
    """Logs messages to file.

    Parameters
    ----------
    filepath
        Path to log file.
    """

    def __init__(self, filepath: str) -> None:
        self.filepath = filepath

    def log(self, msg: str) -> None:
        with open(self.filepath, "a") as f:
            f.write(f"{msg}\n")
